// asm.cpp
// Revision 26-jan-2004

#include "asm.h"

#include <iostream>
using std::cout;
using std::cerr;
using std::endl;

#include <fstream>
#include <iomanip>
using std::dec;
#include <map>
#include <memory>

#include <ctype.h>

#include <assert.h>
#define ASSERT assert

//*********************************************************
//		Auxiliary functions and constants.
//*********************************************************

namespace {

const byte prefixIX= 0xDD;
const byte prefixIY= 0xFD;

std::ostream & hex2 (std::ostream & os)
{
	os << std::hex << std::uppercase <<
		std::setw (2) << std::setfill ('0');
	return os;
}

std::ostream & hex4 (std::ostream & os)
{
	os << std::hex << std::uppercase <<
		std::setw (4) << std::setfill ('0');
	return os;
}

} // namespace

//***********************************************
//		Auxiliary tables
//***********************************************

namespace {

typedef std::map <TypeToken, byte> simple1byte_t;
simple1byte_t simple1byte;
simple1byte_t simpleED;

bool init_maps ()
{
	simple1byte [TypeCCF]=  0x3F;
	simple1byte [TypeCPL]=  0x2F;
	simple1byte [TypeDAA]=  0x27;
	simple1byte [TypeDI]=   0xF3;
	simple1byte [TypeEI]=   0xFB;
	simple1byte [TypeEXX]=  0xD9;
	simple1byte [TypeHALT]= 0x76;
	simple1byte [TypeNOP]=  0x00;
	simple1byte [TypeSCF]=  0x37;
	simple1byte [TypeRLA]=  0x17;
	simple1byte [TypeRLCA]= 0x07;
	simple1byte [TypeRRA]=  0x1F;
	simple1byte [TypeRRCA]= 0x0F;

	simpleED [TypeCPD]=     0xA9;
	simpleED [TypeCPDR]=    0xB9;
	simpleED [TypeCPI]=     0xA1;
	simpleED [TypeCPIR]=    0xB1;
	simpleED [TypeIND]=     0xAA;
	simpleED [TypeINDR]=    0xBA;
	simpleED [TypeINI]=     0xA2;
	simpleED [TypeINIR]=    0xB2;
	simpleED [TypeLDD]=     0xA8;
	simpleED [TypeLDDR]=    0xB8;
	simpleED [TypeLDI]=     0xA0;
	simpleED [TypeLDIR]=    0xB0;
	simpleED [TypeNEG]=     0x44;
	simpleED [TypeOTDR]=    0xBB;
	simpleED [TypeOTIR]=    0xB3;
	simpleED [TypeOUTD]=    0xAB;
	simpleED [TypeOUTI]=    0xA3;
	simpleED [TypeRETI]=    0x4D;
	simpleED [TypeRETN]=    0x45;
	simpleED [TypeRLD]=     0x6F;
	simpleED [TypeRRD]=     0x67;
	return true;
}

bool map_ready= init_maps ();

} // namespace

//*********************************************************
//			class Asm
//*********************************************************

Asm::Asm () :
	base (0),
	current (0),
	minused (65535),
	maxused (0),
	hasentrypoint (false)
{
}

Asm::~Asm ()
{
}

void Asm::setbase (address addr)
{
	base= addr;
	current= addr;
}

void Asm::setentrypoint (address addr)
{
	if (pass < 2)
		return;
	if (hasentrypoint)
		cerr << "WARNING: Entry point redefined" << endl;
	hasentrypoint= true;
	entrypoint= addr;
}

void Asm::checkendline (Tokenizer & tz)
{
	Token tok= tz.gettoken ();
	if (tok.type () != TypeEndLine)
	{
		cerr << "Found " << tok.str () << endl;
		throw "End line expected";
	}
}

void Asm::gencode (byte code)
{
	if (current < minused)
		minused= current;
	if (current > maxused)
		maxused= current;
	mem [current]= code;
	++current;
}

void Asm::gencodeword (address value)
{
	gencode (value & 0xFF);
	gencode (value >> 8);
}

void Asm::setvalue (const std::string & var, address value)
{
	mapvar [var]= value;
}

address Asm::getvalue (const std::string & var)
{
	mapvar_t::iterator it= mapvar.find (var);
	if (it == mapvar.end () )
		throw "Etiqueta o variable sin definir";
	return it->second;
}

address Asm::getvalue (const Token & tok)
{
	switch (tok.type () )
	{
	case TypeNumber:
		return tok.num ();
	case TypeIdentifier:
		return getvalue (tok.str () );
	case TypeDollar:
		return current;
	case TypeLiteral:
		if (tok.str ().size () != 1)
			throw "Invalid literal";
		return tok.str () [0];
	default:
		throw "Value expected";
	}
}

address Asm::getvalue0 (const std::string & var)
{
	mapvar_t::iterator it= mapvar.find (var);
	if (it == mapvar.end () )
	{
		if (pass < 2)
			return 0;
		else
			throw var + " undefined";
	}
	return it->second;
}

address Asm::getvalue0 (const Token & tok)
{
	switch (tok.type () )
	{
	case TypeNumber:
		return tok.num ();
	case TypeIdentifier:
		return getvalue0 (tok.str () );
	case TypeDollar:
		return current;
	default:
		throw "Value Expected";
	}
}

void Asm::parsevalue (bool required, Tokenizer & tz, address & result)
{
	Token tok= tz.gettoken ();
	switch (tok.type () )
	{
	case TypeNumber:
		result= tok.num ();
		break;
	case TypeIdentifier:
		if (required)
			result= getvalue (tok.str () );
		else
			result= getvalue0 (tok.str () );
		break;
	case TypeDollar:
		result= current;
		break;
	case TypeLiteral:
		if (tok.str ().size () != 1)
			throw "Invalid literal";
		result= tok.str () [0];
		break;
	default:
		throw "Value Expected";
	}
}

void Asm::parseopen (bool required, Tokenizer & tz, address & result)
{
	Token tok= tz.gettoken ();
	if (tok.type () == TypeOpen)
	{
		parseorxor (required, tz, result);
		tok= tz.gettoken ();
		if (tok.type () != TypeClose)
			throw "Close expected";	
	}
	else
	{
		tz.putback (tok);
		parsevalue (required, tz, result);
	}
}

void Asm::parsemuldiv (bool required, Tokenizer & tz, address & result)
{
	parseopen (required, tz, result);
	Token tok= tz.gettoken ();
	TypeToken tt= tok.type ();
	while (tt == TypeMult || tt == TypeDiv || tt == TypeMOD ||
		tt == TypeSHL || tt == TypeSHR)
	{
		address guard;
		parseopen (required, tz, guard);
		switch (tt)
		{
		case TypeMult:
			result*= guard;
			break;
		case TypeDiv:
			if (guard == 0)
			{
				if (required || pass >= 2)
					throw "Division by zero";
				else
					result= 0;
			}
			else
				result/= guard;
			break;
		case TypeMOD:
			if (guard == 0)
			{
				if (required || pass >= 2)
					throw "Division by zero";
				else
					result= 0;
			}
			else
				result%= guard;
			break;
		case TypeSHL:
			result<<= guard;
			break;
		case TypeSHR:
			result>>= guard;
			break;
		default:
			throw "Unexpected error";
		}
		tok= tz.gettoken ();
		tt= tok.type ();
	}
	tz.putback (tok);
}

void Asm::parseplusmin (bool required, Tokenizer & tz, address & result)
{
	parsemuldiv (required, tz, result);
	Token tok= tz.gettoken ();
	TypeToken tt= tok.type ();
	while (tt == TypePlus || tt == TypeMinus)
	{
		address guard;
		parsemuldiv (required, tz, guard);
		switch (tt)
		{
		case TypePlus:
			result+= guard;
			break;
		case TypeMinus:
			result-= guard;
			break;
		default:
			throw "Unexpected error";
		}
		tok= tz.gettoken ();
		tt= tok.type ();
	}
	tz.putback (tok);
}

void Asm::parsenot (bool required, Tokenizer & tz, address & result)
{
	Token tok= tz.gettoken ();
	TypeToken tt= tok.type ();
	// NOT and unary + and -.
	if (tt == TypeNOT || tt == TypePlus || tt == TypeMinus)
	{
		parsenot (required, tz, result);
		switch (tt)
		{
		case TypeNOT:
			result= ~ result;
			break;
		case TypePlus:
			break;
		case TypeMinus:
			result= - result;
			break;
		default:
			throw "Unexpected error";
		}
	}
	else
	{
		tz.putback (tok);
		parseplusmin (required, tz, result);
	}
}

void Asm::parseand (bool required, Tokenizer & tz, address & result)
{
	parsenot (required, tz, result);
	Token tok= tz.gettoken ();
	TypeToken tt= tok.type ();
	while (tt == TypeAND)
	{
		address guard;
		parsenot (required, tz, guard);
		result&= guard;
		tok= tz.gettoken ();
		tt= tok.type ();
	}
	tz.putback (tok);
}

void Asm::parseorxor (bool required, Tokenizer & tz, address & result)
{
	parseand (required, tz, result);
	Token tok= tz.gettoken ();
	TypeToken tt= tok.type ();
	while (tt == TypeOR || tt == TypeXOR)
	{
		address guard;
		parseand (required, tz, guard);
		switch (tt)
		{
		case TypeOR:
			result|= guard;
			break;
		case TypeXOR:
			result^= guard;
			break;
		default:
			throw "Unexpected error";
		}
		tok= tz.gettoken ();
		tt= tok.type ();
	}
	tz.putback (tok);
}

address Asm::parseexpr (bool required, const Token & tok, Tokenizer & tz)
{
	tz.putback (tok);
	address result;
	parseorxor (required, tz, result);
	return result;
}

void Asm::parsecomma (Tokenizer & tz)
{
	if  (tz.gettoken ().type () != TypeComma)
		throw "Comma expected";
}

void Asm::parseA (Tokenizer & tz)
{
	if  (tz.gettoken ().type () != TypeA)
		throw "Register A expected";
}

void Asm::parseC (Tokenizer & tz)
{
	if  (tz.gettoken ().type () != TypeC)
		throw "Register C expected";
}

void Asm::parseclose (Tokenizer & tz)
{
	if  (tz.gettoken ().type () != TypeClose)
		throw "Close expected";
}

void Asm::parseopen (Tokenizer & tz)
{
	if  (tz.gettoken ().type () != TypeOpen)
		throw "Open expected";
}

void Asm::loadfile (const std::string & filename)
{
	// Load the file in memory.
	std::ifstream file (filename.c_str () );
	if (! file.is_open () )
		throw "Fichero no encontrado";
	vfileref.push_back (FileRef (filename) );
	size_t filenum= vfileref.size () - 1;
	std::string line;
	size_t linenum= 0;

	try
	{	
		while (std::getline (file, line) )
		{
			++linenum;
	
			Tokenizer tz (line);
			Token tok= tz.gettoken ();
			if (tok.type () == TypeINCLUDE)
			{
				std::string includefile= tz.getincludefile ();
				loadfile (includefile);
			}
			else
			{
				vline.push_back (line);
				vlineinfo.push_back
					//(LineInfo (linenum, filenum) );
					(LineInfo (currentline, filenum) );
			}
		}
	}
	catch (...)
	{
		cerr << "ERROR on line " << linenum <<
			" of file " << filename << endl;
		throw;
	}
}

void Asm::getvalidline ()
{
	for (;;)
	{
		if (currentline >= vline.size () )
			break;
		const std::string & line= vline [currentline];
		if (! line.empty () && line [0] != ';')
			break;
		++currentline;
	}
}

void Asm::parseif (Tokenizer & tz)
{
	address v;
	Token tok= tz.gettoken ();
	v= parseexpr (true, tok, tz);
	checkendline (tz);
	if (v != 0)
		++iflevel;
	else
	{
		size_t ifline= currentline;
		int level= 1;
		for ( ++currentline; currentline < vline.size ();
			++currentline)
		{
			getvalidline ();
			if (currentline >= vline.size () )
				break;
			const std::string & line= vline [currentline];
			Tokenizer tz (line);
			tok= tz.gettoken ();
			TypeToken tt= tok.type ();
			if (tt == TypeIF)
				++level;
			else if (tt == TypeELSE)
			{
				++iflevel;
				break;
			}
			else if (tt == TypeENDIF)
			{
				--level;
				if (level == 0)
					break;
			}
			else
				cerr << "- " << line << endl;
		}
		if (currentline >= vline.size () )
		{
			currentline= ifline;
			throw "IF without ENDIF";
		}
	}
}

void Asm::parseelse ()
{
	if (iflevel == 0)
		throw "ELSE without IF";

	size_t elseline= currentline;
	int level= 1;
	for (++currentline; currentline < vline.size (); ++currentline)
	{
		getvalidline ();
		if (currentline >= vline.size () )
			break;
		const std::string & line= vline [currentline];
		Tokenizer tz (line);
		Token tok= tz.gettoken ();
		TypeToken tt= tok.type ();
		if (tt == TypeIF)
			++level;
		else if (tt == TypeENDIF)
		{
			--level;
			if (level == 0)
				break;
		}
		else
			cerr << "- " << line << endl;
	}
	if (currentline >= vline.size () )
	{
		currentline= elseline;
		throw "ELSE without ENDIF";
	}
	--iflevel;
}

void Asm::parseline (Tokenizer & tz)
{
	Token tok= tz.gettoken ();

	switch (tok.type () )
	{
	case TypeEndLine:
		return;
	case TypeORG:
		parseorg (tz);
		break;
	case TypeIdentifier:
		parselabel (tz, tok.str () );
		break;
	case TypeIF:
		parseif (tz);
		break;
	case TypeELSE:
		checkendline (tz);
		parseelse ();
		break;
	case TypeENDIF:
		checkendline (tz);
		if (iflevel == 0)
			throw "ENDIF without IF";
		--iflevel;
		break;
	case TypePUBLIC:
		parsePUBLIC (tz);
		break;
	default:
		parsegeneric (tz, tok);
	}
}

void Asm::dopass ()
{
	current= base;
	iflevel= 0;
	for (currentline= 0; currentline < vline.size (); ++currentline)
	{
		getvalidline ();
		if (currentline >= vline.size () )
			break;
		const std::string & line= vline [currentline];

		cout << line << endl;

		Tokenizer tz (line);
		parseline (tz);

		#if 0
		Token tok= tz.gettoken ();
		if (tok.type () == TypeEndLine)
			continue;

		switch (tok.type () )
		{
		case TypeORG:
			parseorg (tz);
			break;
		case TypeIdentifier:
			parselabel (tz, tok.str () );
			break;
		case TypeIF:
			parseif (tz);
			break;
		case TypeELSE:
			checkendline (tz);
			parseelse ();
			break;
		case TypeENDIF:
			checkendline (tz);
			if (iflevel == 0)
				throw "ENDIF without IF";
			--iflevel;
			break;
		case TypePUBLIC:
			parsePUBLIC (tz);
			break;
		default:
			parsegeneric (tz, tok);
		}
		#endif

	}
	if (iflevel > 0)
		throw "IF without ENDIF";
}

void Asm::processfile (const std::string & filename)
{
	loadfile (filename);

	// Process the file(s) readed.

	try 
	{
		for (pass= 1; pass <= 2; ++pass)
		{
			dopass ();
		}
	}
	catch (...)
	{
		if (currentline >= vlineinfo.size () )
			throw "Unexpected error after end of file";
		LineInfo & linf= vlineinfo [currentline];
		cerr << dec << "ERROR on line " << linf.linenum <<
			" of file " << vfileref [linf.filenum].filename <<
			endl;
		throw;
	}
}

void Asm::parsegeneric (Tokenizer & tz, Token tok)
{
	TypeToken t= tok.type ();
	simple1byte_t::iterator it= simple1byte.find (t);
	if (it != simple1byte.end () )
	{
		if (tz.gettoken ().type () != TypeEndLine)
			throw "End of line expected";
		gencode (it->second);
		cout << hex2 << int (it->second) << '\t' <<
			tok.str () << endl;
		return;
	}
	it= simpleED.find (t);
	if (it != simpleED.end () )
	{
		if (tz.gettoken ().type () != TypeEndLine)
			throw "End of line expected";
		gencode (0xED);
		gencode (it->second);
		cout << "ED " << hex2 << int (it->second) << '\t' <<
			tok.str () << endl;
		return;
	}

	switch (t)
	{
	case TypeDEFB:
	case TypeDB:
	case TypeDEFM:
		parseDEFB (tz);
		break;
	case TypeDEFW:
	case TypeDW:
		parseDEFW (tz);
		break;
	case TypeDEFS:
	case TypeDS:
		parseDEFS (tz);
		break;
	case TypeEND:
		parseEND (tz);
		break;
	case TypeIM:
		parseIM (tz);
		break;
	case TypeRST:
		parseRST (tz);
		break;
	case TypeLD:
		parseLD (tz);
		break;
	case TypeCP:
		parseCP (tz);
		break;
	case TypeAND:
		parseAND (tz);
		break;
	case TypeOR:
		parseOR (tz);
		break;
	case TypeXOR:
		parseXOR (tz);
		break;
	case TypeRL:
		parseRL (tz);
		break;
	case TypeRLC:
		parseRLC (tz);
		break;
	case TypeRR:
		parseRR (tz);
		break;
	case TypeRRC:
		parseRRC (tz);
		break;
	case TypeSLA:
		parseSLA (tz);
		break;
	case TypeSRA:
		parseSRA (tz);
		break;
	case TypeSRL:
		parseSRL (tz);
		break;
	case TypeSUB:
		parseSUB (tz);
		break;
	case TypeADD:
		parseADD (tz);
		break;
	case TypeADC:
		parseADC (tz);
		break;
	case TypeSBC:
		parseSBC (tz);
		break;
	case TypePUSH:
		parsePUSHPOP (tz, true);
		break;
	case TypePOP:
		parsePUSHPOP (tz, false);
		break;
	case TypeCALL:
		parseCALL (tz);
		break;
	case TypeRET:
		parseRET (tz);
		break;
	case TypeJP:
		parseJP (tz);
		break;
	case TypeJR:
		parseJR (tz);
		break;
	case TypeDJNZ:
		parseDJNZ (tz);
		break;
	case TypeDEC:
		parseINCDEC (tz, false);
		break;
	case TypeINC:
		parseINCDEC (tz, true);
		break;
	case TypeEX:
		parseEX (tz);
		break;
	case TypeIN:
		parseIN (tz);
		break;
	case TypeOUT:
		parseOUT (tz);
		break;
	case TypeBIT:
		parseBIT (tz);
		break;
	case TypeRES:
		parseRES (tz);
		break;
	case TypeSET:
		parseSET (tz);
		break;
	default:
		throw "Syntax error";
	}
}

void Asm::parseorg (Tokenizer & tz)
{
	Token tok= tz.gettoken ();
	address org= parseexpr (true, tok, tz);
	current= org;
	cout << "\tORG " << org << endl;
}

void Asm::parsePUBLIC (Tokenizer & tz)
{
	for (;;)
	{
		Token tok= tz.gettoken ();
		if (tok.type () != TypeIdentifier)
			throw "Invalid PUBLIC declaration";
		setpublic.insert (tok.str () );
		tok= tz.gettoken ();
		if (tok.type () == TypeEndLine)
			break;
		if (tok.type () != TypeComma)
			throw "Invalid PUBLIC declaration";
	}
}

void Asm::parseEND (Tokenizer & tz)
{
	Token tok= tz.gettoken ();
	if (tok.type () != TypeEndLine)
	{
		address end= parseexpr (false, tok, tz);
		checkendline (tz);
		setentrypoint (end);
	}
	currentline= vline.size ();
}

void Asm::parselabel (Tokenizer & tz, const std::string & name)
{
	Token tok= tz.gettoken ();
	if (tok.type () == TypeColon)
		tok= tz.gettoken ();
	if (tok.type () == TypeEQU)
	{
		tok= tz.gettoken ();
		//address value= getvalue (tok);
		address value= parseexpr (false, tok, tz);
		setvalue (name, value);
		checkendline (tz);
		return;
	}

	// En otro caso: etiqueta generica, se le asigna el
	// valor actual del contador de posiciones y se
	// explora el resto de la linea

	setvalue (name, current);
	if (tok.type () != TypeEndLine)
		parsegeneric (tz, tok);
}

enum { regBC= 0, regDE= 1, regHL= 2, regAF= 3, regSP= 3 };

enum { regA= 7, regB= 0, regC= 1, regD= 2, regE= 3,
	regH= 4, regL= 5, reg_HL_= 6
};

void Asm::parsedesp (Tokenizer & tz, byte & desp)
{
	Token tok= tz.gettoken ();
	switch (tok.type () )
	{
	case TypeClose:
		desp= 0;
		break;
	case TypePlus:
		tok= tz.gettoken ();
		{
			address addr= parseexpr (false, tok, tz);
			// We allow positive greater than 127 just in
			// case someone uses hexadecimals such as 0FFh
			// as offsets.
			if (addr > 255)
				throw "Offset out of range";
			desp= static_cast <byte> (addr);
			parseclose (tz);
		}
		break;
	case TypeMinus:
		tok= tz.gettoken ();
		{
			address addr= parseexpr (false, tok, tz);
			if (addr > 128)
				throw "Offset out of range";
			desp= static_cast <byte> (256 - addr);
			parseclose (tz);
		}
		break;
	default:
		throw "Expected '+', '-' or ')'";
	}
}

bool Asm::parsebyteparam (Tokenizer & tz, TypeToken tt,
	unsigned short & regcode,
	byte & prefix, bool & hasdesp, byte & desp)
{
	prefix= 0;
	hasdesp= false;
	desp= 0;
	Token tok;
	switch (tt)
	{
	case TypeA:
		regcode= regA; break;
	case TypeB:
		regcode= regB; break;
	case TypeC:
		regcode= regC; break;
	case TypeD:
		regcode= regD; break;
	case TypeE:
		regcode= regE; break;
	case TypeH:
		regcode= regH; break;
	case TypeL:
		regcode= regL; break;
	case TypeOpen:
		tok= tz.gettoken ();
		switch (tok.type () )
		{
		case TypeHL:
			regcode= reg_HL_;
			parseclose (tz);
			break;
		case TypeIX:
			regcode= reg_HL_;
			prefix= prefixIX;
			hasdesp= true;
			parsedesp (tz, desp);
			break;
		case TypeIY:
			regcode= reg_HL_;
			prefix= prefixIY;
			hasdesp= true;
			parsedesp (tz, desp);
			break;
		default:
			throw "Error por ahora";
		}
		break;
	default:
		return false;
	}
	return true;
}

void Asm::dobyteparam (Tokenizer & tz, byte codereg, byte codein)
{
	Token tok= tz.gettoken ();
	unsigned short reg;
	byte prefix;
	bool hasdesp;
	byte desp;
	if (parsebyteparam (tz, tok.type (), reg, prefix, hasdesp, desp) )
	{
		checkendline (tz);
		if (prefix)
		{
			gencode (prefix);
			cout << hex2 << int (prefix) << ' ';
		}
		byte code= codereg + reg;
		gencode (code);
		cout << hex2 << int (code);
		if (hasdesp)
		{
			gencode (desp);
			cout << ' ' << hex2 << int (desp);
		}
		cout << endl;
	}
	else
	{
		//address addr= getvalue0 (tok);
		address addr= parseexpr (false, tok, tz);
		checkendline (tz);
		gencode (codein);
		byte param= static_cast <byte> (addr);
		gencode (param);
		cout << hex2 << int (codein) << ' ' << hex2 << int (param) <<
			endl;
	}
}

void Asm::dobyteparamCB (Tokenizer & tz, byte codereg)
{
	Token tok= tz.gettoken ();
	unsigned short reg;
	byte prefix;
	bool hasdesp;
	byte desp;
	if (parsebyteparam (tz, tok.type (), reg, prefix, hasdesp, desp) )
	{
		checkendline (tz);
		if (prefix)
		{
			gencode (prefix);
			cout << hex2 << int (prefix) << ' ';
		}
		gencode (0xCB);
		cout << "CB ";
		if (hasdesp)
		{
			gencode (desp);
			cout << hex2 << int (desp) << ' ';
		}
		byte code= codereg + reg;
		gencode (code);
		cout << hex2 << int (code);
		cout << endl;
	}
	else
		throw "Invalid operand";
}

void Asm::parseIM (Tokenizer & tz)
{
	Token tok= tz.gettoken ();
	//address v= getvalue (tok);
	address v= parseexpr (true, tok, tz);
	byte code;
	switch (v)
	{
	case 0:
		code= 0x46; break;
	case 1:
		code= 0x56; break;
	case 2:
		code= 0x5E; break;
	default:
		throw "Invalid IM value";
	}
	checkendline (tz);
	gencode (0xED);
	gencode (code);
	cout << "ED " << hex2 << int (code) <<
		"\tIM " << v << endl;
}

void Asm::parseRST (Tokenizer & tz)
{
	Token tok= tz.gettoken ();
	//address v= getvalue (tok);
	address v= parseexpr (true, tok, tz);
	byte code;
	switch (v)
	{
	case 0x00:
		code= 0xC7; break;
	case 0x08:
		code= 0xCF; break;
	case 0x10:
		code= 0xD7; break;
	case 0x18:
		code= 0xDF; break;
	case 0x20:
		code= 0xE7; break;
	case 0x28:
		code= 0xEF; break;
	case 0x30:
		code= 0xF7; break;
	case 0x38:
		code= 0xFF; break;
	default:
		throw "Invalid RST value";
	}
	checkendline (tz);
	gencode (code);
	cout << int (code) <<
		"\tRST " << hex2 << v << endl;
}

void Asm::parseLDA (Tokenizer & tz)
{
	parsecomma (tz);
	Token tok= tz.gettoken ();
	switch (tok.type () )
	{
	case TypeA:
		gencode (0x7F);
		cout << "7F\tLD A, A" << endl;
		break;
	case TypeB:
		gencode (0x78);
		cout << "78\tLD A, B" << endl;
		break;
	case TypeC:
		gencode (0x79);
		cout << "79\tLD A, C" << endl;
		break;
	case TypeD:
		gencode (0x7A);
		cout << "7A\tLD A, D" << endl;
		break;
	case TypeE:
		gencode (0x7B);
		cout << "7B\tLD A, E" << endl;
		break;
	case TypeH:
		gencode (0x7C);
		cout << "7C\tLD A, H" << endl;
		break;
	case TypeI:
		gencode (0xED);
		gencode (0x57);
		cout << "ED 57\tLD A, I" << endl;
		break;
	case TypeL:
		gencode (0x7D);
		cout << "7D\tLD A, L" << endl;
		break;
	case TypeR:
		gencode (0xED);
		gencode (0x5F);
		cout << "ED 5F\tLD A, R" << endl;
		break;
	case TypeOpen:
		tok= tz.gettoken ();
		switch (tok.type () )
		{
		case TypeBC:
			parseclose (tz);
			gencode (0x0A);
			cout << "0A\tLD A, (BC)" << endl;
			break;
		case TypeDE:
			parseclose (tz);
			gencode (0x1A);
			cout << "1A\tLD A, (DE)" << endl;
			break;
		case TypeHL:
			parseclose (tz);
			gencode (0x7E);
			cout << "7E\tLD A, (HL)" << endl;
			break;
		case TypeIX:
			{
				byte desp;
				parsedesp (tz, desp);
				gencode (prefixIX);
				gencode (0x7E);
				gencode (desp);
				cout << "DD 7E " <<
					hex2 << int (desp) <<
					"\tLD A, (IX+n)" << endl;
			}
			break;
		case TypeIY:
			{
				byte desp;
				parsedesp (tz, desp);
				gencode (prefixIY);
				gencode (0x7E);
				gencode (desp);
				cout << "DD 7E " <<
					hex2 << int (desp) <<
					"\tLD A, (IX+n)" << endl;
			}
			break;
		default:
			{
				//address addr= getvalue0 (tok);
				address addr= parseexpr (false, tok, tz);
				parseclose (tz);
				gencode (0x3A);
				gencodeword (addr);
				cout << "3A " << hex2 << (addr & 0xFF) <<
					' ' << (addr >> 8) <<
					"\tLD A, (nn)" << endl;
			}
		}
		break;
	default:
		//address value= getvalue (tok);
		address value= parseexpr (true, tok, tz);
		gencode (0x3E);
		gencode (value & 0xFF);
		cout << "3E " << hex2 << (value & 0xFF) <<
			"\tLD A, n" << endl;
	}
	checkendline (tz);
}
void Asm::parseLDsimple (Tokenizer & tz, unsigned short regcode)
{
	#if 0
	switch (regcode)
	{
	case regB:
		code= 0x06;
		break;
	case regC:
		code= 0x0E;
		break;
	case regD:
		code= 0x16;
		break;
	case regE:
		code= 0x1E;
		break;
	case regH:
		code= 0x26;
		break;
	case regL:
		code= 0x2E;
		break;
	case reg_HL_:
		code= 0x36;
		break;
	default:
		throw "Error por ahora";
	}
	#endif

	parsecomma (tz);
	Token tok= tz.gettoken ();

	unsigned short reg2;
	#if 0
	switch (tok.type () )
	{
	case TypeA:
		reg2= regA; break;
	case TypeB:
		reg2= regB; break;
	case TypeC:
		reg2= regC; break;
	case TypeD:
		reg2= regD; break;
	case TypeE:
		reg2= regE; break;
	case TypeH:
		reg2= regH; break;
	case TypeL:
		reg2= regL; break;
	case TypeOpen:
		tok= tz.gettoken ();
		switch (tok.type () )
		{
		case TypeHL:
			reg2= reg_HL_;
			parseclose (tz);
			break;
		default:
			throw "Error por ahora";
		}
		break;
	default:
		checkendline (tz);
		// Por ahora
		code= (regcode << 3) + 0x06;
		//address value= getvalue (tok);
		address value= parseexpr (true, tok, tz);

		gencode (code);
		gencode (value & 0xFF);
		cout << hex2 << int (code) << ' ' <<
			hex2 << (value & 0xFF) << endl;
		return;
	}
	#else
	byte prefix;
	bool hasdesp;
	byte desp;
	if (parsebyteparam (tz, tok.type (), reg2, prefix, hasdesp, desp) )
	{
		checkendline (tz);
		if (prefix)
		{
			gencode (prefix);
			cout << hex2 << int (prefix) << ' ';
		}
		byte code= 0x40 + (regcode << 3) + reg2;
		gencode (code);
		cout << hex2 << int (code);
		if (hasdesp)
		{
			gencode (desp);
			cout << ' ' << hex2 << int (desp);
		}
		cout << endl;
	}
	else
	{
		// Por ahora
		byte code= (regcode << 3) + 0x06;
		//address value= getvalue (tok);
		address value= parseexpr (true, tok, tz);
		checkendline (tz);
		gencode (code);
		gencode (value & 0xFF);
		cout << hex2 << int (code) << ' ' <<
			hex2 << (value & 0xFF) << endl;
	}
	#endif
}

void Asm::parseLDdouble (Tokenizer & tz, unsigned short regcode, byte prefix)
{
	parsecomma (tz);
	Token tok= tz.gettoken ();

	if (tok.type () == TypeOpen)
	{
		tok= tz.gettoken ();
		//address value= getvalue0 (tok);
		address value= parseexpr (false, tok, tz);
		parseclose (tz);
		checkendline (tz);
		switch (regcode)
		{
		case regBC:
			gencode (0xED);
			gencode (0x4B);
			gencodeword (value);
			cout << "ED 4B " << hex4 << value <<
				"\tLD BC, (nn)" << endl;
			break;
		case regDE:
			gencode (0xED);
			gencode (0x5B);
			gencodeword (value);
			cout << "ED 5B " << hex4 << value <<
				"\tLD DE, (nn)" << endl;
			break;
		case regHL:
			if (prefix)
			{
				gencode (prefix);
				cout << hex2 << int (prefix) << ' ';
			}
			gencode (0x2A);
			gencodeword (value);
			cout << "2A " << hex4 << value << endl;
			break;
		default:
			throw "Unexpected error";
		}
		return;
	}

	// Por ahora
	//address value= getvalue0 (tok);
	address value= parseexpr (false, tok, tz);
	checkendline (tz);

	if (prefix)
	{
		gencode (prefix);
		cout << hex2 << int (prefix) << ' ';
	}
	gencode (regcode * 16 + 1);
	gencodeword (value);
	cout << hex2 << regcode * 16 + 1 << ' ' <<
		hex4 << value << endl;
}

void Asm::parseLDSP (Tokenizer & tz)
{
	parsecomma (tz);
	Token tok= tz.gettoken ();
	switch (tok.type () )
	{
	case TypeHL:
		gencode (0xF9);
		cout << "F9\tLD SP, HL" << endl;
		break;
	case TypeIX:
		gencode (prefixIX);
		gencode (0xF9);
		cout << "DD F9\tLD SP, IX" << endl;
		break;
	case TypeIY:
		gencode (prefixIY);
		gencode (0xF9);
		cout << "FD F9\tLD SP, IY" << endl;
		break;
	case TypeOpen:
		{
			tok= tz.gettoken ();
			//address addr= getvalue0 (tok);
			address addr= parseexpr (false, tok, tz);
			parseclose (tz);
			gencode (0xED);
			gencode (0x7B);
			gencodeword (addr);
			cout << "ED 7B " << hex4 << addr <<
				"\tLD SP, (nn)" << endl;
		}
		break;
	default:
		{
			//address addr= getvalue0 (tok);
			address addr= parseexpr (false, tok, tz);
			gencode (0x31);
			gencodeword (addr);
			cout << "31 " << hex4 << addr <<
				"\tLD SP, nn" << endl;
		}
	}
	checkendline (tz);
}

void Asm::parseLD (Tokenizer & tz)
{
	Token tok= tz.gettoken ();
	switch (tok.type () )
	{
	case TypeA:
		//parseLDsimple (tz, regA);
		parseLDA (tz);
		break;
	case TypeB:
		parseLDsimple (tz, regB);
		break;
	case TypeC:
		parseLDsimple (tz, regC);
		break;
	case TypeD:
		parseLDsimple (tz, regD);
		break;
	case TypeE:
		parseLDsimple (tz, regE);
		break;
	case TypeH:
		parseLDsimple (tz, regH);
		break;
	case TypeL:
		parseLDsimple (tz, regL);
		break;
	case TypeI:
		parsecomma (tz);
		parseA (tz);
		checkendline (tz);
		gencode (0xED);
		gencode (0x47);
		cout << "ED 47\tLD I, A" << endl;
		break;
	case TypeR:
		parsecomma (tz);
		parseA (tz);
		checkendline (tz);
		gencode (0xED);
		gencode (0x4F);
		cout << "ED 4F\tLD R, A" << endl;
		break;
	case TypeBC:
		parseLDdouble (tz, regBC, 0);
		break;
	case TypeDE:
		parseLDdouble (tz, regDE, 0);
		break;
	case TypeHL:
		parseLDdouble (tz, regHL, 0);
		break;
	case TypeIX:
		parseLDdouble (tz, regHL, prefixIX);
		break;
	case TypeIY:
		parseLDdouble (tz, regHL, prefixIY);
		break;
	case TypeSP:
		parseLDSP (tz);
		break;
	case TypeOpen:
		tok= tz.gettoken ();
		switch (tok.type () )
		{
		case TypeBC:
			parseclose (tz);
			parsecomma (tz);
			tok= tz.gettoken ();
			if (tok.type () != TypeA)
				throw "Invalid operand";
			checkendline (tz);
			gencode (0x02);
			cout << "02\tLD (BC), A" << endl;
			break;
		case TypeDE:
			parseclose (tz);
			parsecomma (tz);
			tok= tz.gettoken ();
			if (tok.type () != TypeA)
				throw "Invalid operand";
			checkendline (tz);
			gencode (0x12);
			cout << "12\tLD (DE), A" << endl;
			break;
		case TypeHL:
			parseclose (tz);
			parseLDsimple (tz, reg_HL_);
			break;
		case TypeIX:
			{
				byte desp;
				parsedesp (tz, desp);
				parsecomma (tz);
				tok= tz.gettoken ();
				unsigned short reg;
				byte prefix= 0;
				bool hasdesp;
				byte despnotused;
				if (parsebyteparam (tz, tok.type (),
					reg, prefix, hasdesp, despnotused) )
				{
					checkendline (tz);
					if (prefix || hasdesp ||
							reg == reg_HL_)
						throw "Operand invalid";
					byte code= 0x70 + reg;
					gencode (prefixIX);
					gencode (code);
					gencode (desp);
					cout << "DD "  <<
						hex2 << int (code) <<
						' ' << hex2 << int (desp) <<
						endl;
				}
				else
				{
					//address addr= getvalue0 (tok);
					address addr=
						parseexpr (false, tok, tz);
					checkendline (tz);
					gencode (prefixIX);
					gencode (0x36);
					gencode (desp);
					byte n= static_cast <byte> (addr);
					gencode (n);
					cout << "DD 36 " <<
						hex2 << int (desp) << ' ' <<
						hex2 << int (n) << endl;
				}
			}
			break;
		case TypeIY:
			{
				byte desp;
				parsedesp (tz, desp);
				parsecomma (tz);
				tok= tz.gettoken ();
				unsigned short reg;
				byte prefix= 0;
				bool hasdesp;
				byte despnotused;
				if (parsebyteparam (tz, tok.type (),
					reg, prefix, hasdesp, despnotused) )
				{
					checkendline (tz);
					if (prefix || hasdesp ||
							reg == reg_HL_)
						throw "Operand invalid";
					byte code= 0x70 + reg;
					gencode (prefixIY);
					gencode (code);
					gencode (desp);
					cout << "FD "  <<
						hex2 << int (code) <<
						' ' << hex2 << int (desp) <<
						endl;
				}
				else
				{
					//address addr= getvalue0 (tok);
					address addr=
						parseexpr (false, tok, tz);
					checkendline (tz);
					gencode (prefixIY);
					gencode (0x36);
					gencode (desp);
					byte n= static_cast <byte> (addr);
					gencode (n);
					cout << "FD 36 " <<
						hex2 << int (desp) << ' ' <<
						hex2 << int (n) << endl;
				}
			}
			break;
		default:
			// LD (nn), ...
			{
				//address addr= getvalue0 (tok);
				address addr= parseexpr (false, tok, tz);
				parseclose (tz);
				parsecomma (tz);
				tok= tz.gettoken ();
				byte code;
				byte prefix= 0;
				switch (tok.type () )
				{
				case TypeA:
					code= 0x32;
					break;
				case TypeBC:
					prefix= 0xED;
					code= 0x43;
					break;
				case TypeDE:
					prefix= 0xED;
					code= 0x53;
					break;
				case TypeHL:
					code= 0x22;
					break;
				case TypeIX:
					prefix= prefixIX;
					code= 0x22;
					break;
				case TypeIY:
					prefix= prefixIY;
					code= 0x22;
					break;
				case TypeSP:
					prefix= 0xED;
					code= 0x73;
					break;
				default:
					throw "Operand invalid";
				}
				checkendline (tz);
				if (prefix)
				{
					gencode (prefix);
					cout << hex2 << int (prefix) << ' ';
				}
				gencode (code);
				cout << hex2 << int (code) << ' ';
				gencodeword (addr);
				cout << hex4 << addr << endl;
			}
		}
		break;
	default:
		throw "Operand invalid";
	}
}

void Asm::parseCP (Tokenizer & tz)
{
	#if 0
	unsigned short reg;
	Token tok= tz.gettoken ();
	byte prefix;
	bool hasdesp;
	byte desp;
	if (parsebyteparam (tz, tok.type (), reg, prefix, hasdesp, desp) )
	{
		checkendline (tz);
		if (prefix)
		{
			gencode (prefix);
			cout << hex2 << int (prefix) << ' ';
		}
		byte code= 0xB8 + reg;
		gencode (code);
		cout << hex2 << int (code);
		if (hasdesp)
		{
			gencode (desp);
			cout << ' ' << hex2 << int (desp);
		}
		cout << endl;
	}
	else
	{
		//address addr= getvalue0 (tok);
		address addr= parseexpr (false, tok, tz);
		checkendline (tz);
		gencode (0xFE);
		byte param= static_cast <byte> (addr);
		gencode (param);
		cout << "FE " << hex2 << int (param) <<
			"\tCP nn" << endl;
	}
	#else
	dobyteparam (tz, 0xB8, 0xFE);
	#endif
}

void Asm::parseAND (Tokenizer & tz)
{
	#if 0
	unsigned short reg;
	Token tok= tz.gettoken ();
	byte prefix;
	bool hasdesp;
	byte desp;
	if (parsebyteparam (tz, tok.type (), reg, prefix, hasdesp, desp) )
	{
		checkendline (tz);
		if (prefix)
		{
			gencode (prefix);
			cout << hex2 << int (prefix) << ' ';
		}
		byte code= 0xA0 + reg;
		gencode (code);
		cout << hex2 << int (code);
		if (hasdesp)
		{
			gencode (desp);
			cout << ' ' << hex2 << int (desp);
		}
		cout << endl;
	}
	else
	{
		//address addr= getvalue0 (tok);
		address addr= parseexpr (false, tok, tz);
		checkendline (tz);
		gencode (0xE6);
		byte param= static_cast <byte> (addr);
		gencode (param);
		cout << "E6 " << hex2 << int (param) <<
			"\tAND n" << endl;
	}
	#else
	dobyteparam (tz, 0xA0, 0xE6);
	#endif
}

void Asm::parseOR (Tokenizer & tz)
{
	#if 0
	unsigned short reg;
	Token tok= tz.gettoken ();
	byte prefix;
	bool hasdesp;
	byte desp;
	if (parsebyteparam (tz, tok.type (), reg, prefix, hasdesp, desp) )
	{
		checkendline (tz);
		if (prefix)
		{
			gencode (prefix);
			cout << hex2 << int (prefix) << ' ';
		}
		byte code= 0xB0 + reg;
		gencode (code);
		cout << hex2 << int (code);
		if (hasdesp)
		{
			gencode (desp);
			cout << ' ' << hex2 << int (desp);
		}
		cout << endl;
	}
	else
	{
		//address addr= getvalue0 (tok);
		address addr= parseexpr (false, tok, tz);
		checkendline (tz);
		gencode (0xF6);
		byte param= static_cast <byte> (addr);
		gencode (param);
		cout << "F6 " << hex2 << int (param) <<
			"\tOR n" << endl;
	}
	#else
	dobyteparam (tz, 0xB0, 0xF6);
	#endif
}

void Asm::parseXOR (Tokenizer & tz)
{
	dobyteparam (tz, 0xA8, 0xEE);
}

void Asm::parseRL (Tokenizer & tz)
{
	dobyteparamCB (tz, 0x10);
}

void Asm::parseRLC (Tokenizer & tz)
{
	dobyteparamCB (tz, 0x00);
}

void Asm::parseRR (Tokenizer & tz)
{
	dobyteparamCB (tz, 0x18);
}

void Asm::parseRRC (Tokenizer & tz)
{
	dobyteparamCB (tz, 0x08);
}

void Asm::parseSLA (Tokenizer & tz)
{
	dobyteparamCB (tz, 0x20);
}

void Asm::parseSRA (Tokenizer & tz)
{
	dobyteparamCB (tz, 0x28);
}

void Asm::parseSRL (Tokenizer & tz)
{
	dobyteparamCB (tz, 0x38);
}

void Asm::parseSUB (Tokenizer & tz)
{
	dobyteparam (tz, 0x90, 0xD6);
}

void Asm::parseADDADCSBCHL (Tokenizer & tz, byte prefix, byte basecode)
{
	parsecomma (tz);
	Token tok= tz.gettoken ();
	unsigned short reg;
	switch (tok.type () )
	{
	case TypeBC:
		reg= regBC; break;
	case TypeDE:
		reg= regDE; break;
	case TypeHL:
		if (prefix != 0)
			throw "Operand invalid";
		reg= regHL; break;
	case TypeSP:
		reg= regSP; break;
	case TypeIX:
		if (prefix != prefixIX)
			throw "Operand invalid";
		reg= regHL;
		break;
	case TypeIY:
		if (prefix != prefixIY)
			throw "Operand invalid";
		reg= regHL;
		break;
	default:
		throw "Operand invalid";
	}
	checkendline (tz);
	if (prefix)
	{
		gencode (prefix);
		cout << hex2 << int (prefix) << ' ';
	}
	if (basecode == 0x42 || basecode == 0x4A)
	{
		gencode (0xED);
		cout << "ED ";
	}
	//byte code= (reg << 4) + (isADD ? 0x09 : 0x4A);
	byte code= (reg << 4) + basecode;
	gencode (code);
	cout << hex2 << int (code) << endl;
}

void Asm::parseADD (Tokenizer & tz)
{
	//unsigned short reg;
	Token tok= tz.gettoken ();
	//byte prefix;
	//bool hasdesp;
	//byte desp;
	switch (tok.type () )
	{
	case TypeA:
		break;
	case TypeHL:
		parseADDADCSBCHL (tz, 0, 0x09);
		return;
	case TypeIX:
		parseADDADCSBCHL (tz, prefixIX, 0x09);
		return;
	case TypeIY:
		parseADDADCSBCHL (tz, prefixIY, 0x09);
		return;
	default:
		throw "Invalid operand";
	}
	parsecomma (tz);
	#if 0
	tok= tz.gettoken ();
	if (parsebyteparam (tz, tok.type (), reg, prefix, hasdesp, desp) )
	{
		checkendline (tz);
		if (prefix)
		{
			gencode (prefix);
			cout << hex2 << int (prefix) << ' ';
		}
		byte code= 0x80 + reg;
		gencode (code);
		cout << hex2 << int (code);
		if (hasdesp)
		{
			gencode (desp);
			cout << ' ' << hex2 << int (desp);
		}
		cout << endl;
	}
	else
	{
		//address addr= getvalue0 (tok);
		address addr= parseexpr (false, tok, tz);
		checkendline (tz);
		gencode (0xC6);
		byte param= static_cast <byte> (addr);
		gencode (param);
		cout << "C6 " << hex2 << int (param) <<
			"\tADD A, n" << endl;
	}
	#else
	dobyteparam (tz, 0x80, 0xC6);
	#endif
}

void Asm::parseADC (Tokenizer & tz)
{
	Token tok= tz.gettoken ();
	switch (tok.type () )
	{
	case TypeA:
		break;
	case TypeHL:
		parseADDADCSBCHL (tz, 0, 0x4A);
		return;
	default:
		throw "Invalid param to ADC";
	}
	parsecomma (tz);
	#if 0
	tok= tz.gettoken ();
	unsigned short reg;
	byte prefix;
	bool hasdesp;
	byte desp;
	if (parsebyteparam (tz, tok.type (), reg, prefix, hasdesp, desp) )
	{
		checkendline (tz);
		if (prefix)
		{
			gencode (prefix);
			cout << hex2 << int (prefix) << ' ';
		}
		byte code= 0x88 + reg;
		gencode (code);
		cout << hex2 << int (code);
		if (hasdesp)
		{
			gencode (desp);
			cout << ' ' << hex2 << int (desp);
		}
		cout << endl;
	}
	else
	{
		//address addr= getvalue0 (tok);
		address addr= parseexpr (false, tok, tz);
		checkendline (tz);
		gencode (0xCE);
		byte param= static_cast <byte> (addr);
		gencode (param);
		cout << "CE " << hex2 << int (param) <<
			"\tADC A, n" << endl;
	}
	#else
	dobyteparam (tz, 0x88, 0xCE);
	#endif
}

void Asm::parseSBC (Tokenizer & tz)
{
	Token tok= tz.gettoken ();
	switch (tok.type () )
	{
	case TypeA:
		#if 0
		{
			parsecomma (tz);
			tok= tz.gettoken ();
			unsigned short reg;
			byte prefix;
			bool hasdesp;
			byte desp;
			if (parsebyteparam (tz, tok.type (), reg, prefix,
				hasdesp, desp) )
			{
				checkendline (tz);
				if (prefix)
				{
					gencode (prefix);
					cout << hex2 << int (prefix) << ' ';
				}
				byte code= 0x98 + reg;
				gencode (code);
				cout << hex2 << int (code);
				if (hasdesp)
				{
					gencode (desp);
					cout << ' ' << hex2 << int (desp);
				}
				cout << endl;
			}
			else
			{
				//address addr= getvalue0 (tok);
				address addr= parseexpr (false, tok, tz);
				checkendline (tz);
				gencode (0xDE);
				byte param= static_cast <byte> (addr);
				gencode (param);
				cout << "DE " << hex2 << int (param) <<
					"\tSBC A, n" << endl;
			}
		}
		#else
		parsecomma (tz);
		dobyteparam (tz, 0x98, 0xDE);
		#endif
		break;
	case TypeHL:
		parseADDADCSBCHL (tz, 0, 0x42);
		break;
	default:
		throw "Invalid operand";
	}
}

// Casos
//	push bc -> C5
//	push de -> D5
//	push hl -> E5
//	push af -> F5
//	push ix -> DD E5
//	push iy -> FD E5
//	pop bc -> C1
//	pop de -> D1
//	pop hl -> E1
//	pop af -> F1
//	pop ix -> DD E1
//	pop iy -> FD E1

void Asm::parsePUSHPOP (Tokenizer & tz, bool isPUSH)
{
	Token tok= tz.gettoken ();
	byte code= 0;
	byte prefix= 0;

	//cerr << "Tipo: " << tok.type () << endl;
	switch (tok.type () )
	{
	case TypeBC:
		code= regBC;
		break;
	case TypeDE:
		code= regDE;
		break;
	case TypeHL:
		code= regHL;
		break;
	case TypeAF:
		code= regAF;
		break;
	case TypeIX:
		code= regHL;
		prefix= prefixIX;
		break;
	case TypeIY:
		code= regHL;
		prefix= prefixIY;
		break;
	default:
		throw "Error de sintaxis";
	}
	checkendline (tz);

	//cerr << "Code= " << int (code) << endl;
	code<<= 4;
	code+= isPUSH ? 0xC5 : 0xC1;

	if (prefix)
		gencode (prefix);
	gencode (code);

	if (prefix)
		cout << hex2 << int (prefix) << ' ';
	cout << hex2 << int (code) << endl;
}

// Codigos JP y CALL

// jp NN     -> C3
// jp (hl)   -> E9
// jp nz, NN -> C2
// jp z,  NN -> CA
// jp nc, NN -> D2
// jp c, NN  -> DA
// jp po, NN -> E2
// jp pe, NN -> EA
// jp p, NN  -> F2
// jp m, NN  -> FA

// call NN     -> CD
// call nz, NN -> C4
// call z, NN  -> CC
// call nc, NN -> D4
// call c, NN  -> DC
// call po, NN -> E4
// call pe, NN -> EC
// call p, NN  -> F4
// call m, NN  -> FC

void Asm::parseCALL (Tokenizer & tz)
{
	Token tok= tz.gettoken ();
	address addr;
	#if 0
	switch (tok.type () )
	{
	case TypeNumber:
		addr= tok.num ();
		break;
	case TypeIdentifier:
		//addr= getvalue0 (tok.str () );
		addr= parseexpr (false, tok, tz);
		break;
	default:
		throw "Error de sintaxis";
	}
	#else
	byte code;
	switch (tok.type () )
	{
	case TypeC:
		code= 0xDC; break;
	case TypeM:
		code= 0xFC; break;
	case TypeNC:
		code= 0xD4; break;
	case TypeNZ:
		code= 0xC4; break;
	case TypeP:
		code= 0xF4; break;
	case TypePE:
		code= 0xEC; break;
	case TypePO:
		code= 0xE4; break;
	case TypeZ:
		code= 0xCC; break;
	default:
		code= 0xCD;
	}
	if (code != 0xCD)
	{
		parsecomma (tz);
		tok= tz.gettoken ();
	}
	//addr= getvalue0 (tok);
	addr= parseexpr (false, tok, tz);
	checkendline (tz);
	#endif
	gencode (code);
	gencodeword (addr);
	cout << hex2 << int (code) << ' ' << hex4 << addr << endl;
}

void Asm::parseRET (Tokenizer & tz)
{
	Token tok= tz.gettoken ();
	// Provisional
	byte code= 0;
	switch (tok.type () )
	{
	case TypeEndLine:
		code= 0xC9;
		break;
	case TypeNZ:
		code= 0xC0;
		break;
	case TypeZ:
		code= 0xC8;
		break;
	case TypeNC:
		code= 0xD0;
		break;
	case TypeC:
		code= 0xD8;
		break;
	case TypePO:
		code= 0xE0;
		break;
	case TypePE:
		code= 0xE8;
		break;
	case TypeP:
		code= 0xF0;
		break;
	case TypeM:
		code= 0xF8;
		break;
	default:
		throw "Sytax error";
	}
	if (code != 0xC9)
		checkendline (tz);
	gencode (code);
	cout << hex2 << int (code) << "\tRET" << endl;
}

void Asm::parseJP (Tokenizer & tz)
{
	Token tok= tz.gettoken ();
	byte code= 0;
	switch (tok.type () )
	{
	case TypeNZ:
		code= 0xC2;
		break;
	case TypeZ:
		code= 0xCA;
		break;
	case TypeNC:
		code= 0xD2;
		break;
	case TypeC:
		code= 0xDA;
		break;
	case TypePO:
		code= 0xE2;
		break;
	case TypePE:
		code= 0xEA;
		break;
	case TypeP:
		code= 0xF2;
		break;
	case TypeM:
		code= 0xFA;
		break;
	case TypeOpen:
		tok= tz.gettoken ();
		{
			byte prefix= 0;
			switch (tok.type () )
			{
			case TypeHL:
				break;
			case TypeIX:
				prefix= prefixIX;
				break;
			case TypeIY:
				prefix= prefixIY;
				break;
			default:
				throw "Invalid JP ()";
			}
			parseclose (tz);
			checkendline (tz);
			if (prefix)
			{
				gencode (prefix);
				cout << hex2 << int (prefix) << ' ';
			}
			gencode (0xE9);
			cout << "E9\tJP ()" << endl;
		}
		return;
	default:
		code= 0xC3;
	}
	if (code != 0xC3)
	{
		tok= tz.gettoken ();
		if (tok.type () != TypeComma)
			throw "Esperada coma";
		tok= tz.gettoken ();
	}

	address addr= 0;
	switch (tok.type () )
	{
	case TypeNumber:
		addr= tok.num ();
		break;
	case TypeIdentifier:
		//addr= getvalue0 (tok.str () );
		addr= parseexpr (false, tok, tz);
		break;
	default:
		throw "Error de sintaxis";
	}
	checkendline (tz);

	gencode (code);
	gencodeword (addr);
	cout << hex2 << int (code) << ' ' << hex4 << addr << endl;
}

void Asm::parseJR (Tokenizer & tz)
{
	Token tok= tz.gettoken ();
	byte code= 0;
	switch (tok.type () )
	{
	case TypeNZ:
		code= 0x20;
		break;
	case TypeZ:
		code= 0x28;
		break;
	case TypeNC:
		code= 0x30;
		break;
	case TypeC:
		code= 0x38;
		break;
	default:
		code= 0x18;
	}
	if (code != 0x18)
	{
		tok= tz.gettoken ();
		if (tok.type () != TypeComma)
			throw "Esperada coma";
		tok= tz.gettoken ();
	}
	//bool defined= true;
	#if 0
	address addr= 0;
	switch (tok.type () )
	{
	case TypeNumber:
		addr= tok.num ();
		break;
	case TypeIdentifier:
		//addr= getvalue0 (tok.str () );
		addr= parseexpr (false, tok, tz);
		break;
	default:
		throw "Error de sintaxis";
	}
	#else
	address addr= parseexpr (false, tok, tz);
	#endif
	int dif= 0;
	//if (defined)
	if (pass >= 2)
	{
		dif= addr - (current + 2);
		if (dif > 127 || dif < -128)
		{
			cout << dec << "addr= " << addr <<
				" current= " << current <<
				" dif= " << dif << endl;
			throw "JR out of range";
		}
	}
	gencode (code);
	signed char c= static_cast <signed char> (dif);
	gencode (c);
	cout << hex2 << int (code) << ' ' <<
		hex2 << (int (c) & 0xFF) << endl;
}

void Asm::parseDJNZ (Tokenizer & tz)
{
	Token tok= tz.gettoken ();
	//bool defined= true;
	#if 0
	address addr= 0;
	switch (tok.type () )
	{
	case TypeNumber:
		addr= tok.num ();
		break;
	case TypeIdentifier:
		#if 0
		defined= valuedefined (tok.str () );
		if (defined)
			addr= getvalue (tok.str () );
		else
			throw "Forward JR unimplemnted yet";
		#else
		addr= getvalue0 (tok.str () );
		#endif
		break;
	default:
		throw "Error de sintaxis";
	}
	#else
	address addr= parseexpr (false, tok, tz);
	#endif
	int dif= 0;
	//if (defined)
	if (pass >= 2)
	{
		dif= addr - (current + 2);
		if (dif > 127 || dif < -128)
			throw "JR out of range";
	}
	gencode (0x10);
	signed char c= static_cast <signed char> (dif);
	gencode (c);
	cout << "10 " << hex2 << (int (c) & 0xFF) << endl;
}

void Asm::parseINCDEC (Tokenizer & tz, bool isINC)
{
	Token tok= tz.gettoken ();
	byte code;
	byte prefix= 0;
	bool hasdesp= false;
	byte desp;
	switch (tok.type () )
	{
	case TypeA:
		code= isINC ? 0x3C : 0x3D;
		break;
	case TypeB:
		code= isINC ? 0x04 : 0x05;
		break;
	case TypeC:
		code= isINC ? 0x0C : 0x0D;
		break;
	case TypeD:
		code= isINC ? 0x14 : 0x15;
		break;
	case TypeE:
		code= isINC ? 0x1C : 0x1D;
		break;
	case TypeH:
		code= isINC ? 0x24 : 0x25;
		break;
	case TypeL:
		code= isINC ? 0x2C : 0x2D;
		break;
	case TypeBC:
		code= isINC ? 0x03 : 0x0B;
		break;
	case TypeDE:
		code= isINC ? 0x13 : 0x1B;
		break;
	case TypeHL:
		code= isINC ? 0x23 : 0x2B;
		break;
	case TypeIX:
		code= isINC ? 0x23 : 0x2B;
		prefix= prefixIX;
		break;
	case TypeIY:
		code= isINC ? 0x23 : 0x2B;
		prefix= prefixIY;
		break;
	case TypeSP:
		code= isINC ? 0x33 : 0x3B;
		break;
	case TypeOpen:
		tok= tz.gettoken ();
		switch (tok.type () )
		{
		case TypeHL:
			parseclose (tz);
			code= isINC ? 0x34 : 0x35;
			break;
		case TypeIX:
			code= isINC ? 0x34 : 0x35;
			prefix= prefixIX;
			hasdesp= true;
			parsedesp (tz, desp);
			break;
		case TypeIY:
			code= isINC ? 0x34 : 0x35;
			prefix= prefixIY;
			hasdesp= true;
			parsedesp (tz, desp);
			break;
		default:
			throw "Invalid operand";
		}
		break;
	default:
		throw "Error por ahora";
	}
	checkendline (tz);

	if (prefix != 0)
		gencode (prefix);
	gencode (code);
	if (hasdesp)
		gencode (desp);

	if (prefix != 0)
		cout << hex2 << int (prefix) <<  ' ';
	cout << hex2 << int (code);
	if (hasdesp)
		cout << ' ' << hex2 << int (desp);
	cout << '\t' <<
		(isINC ? "INC" : "DEC") << endl;
}

void Asm::parseEX (Tokenizer & tz)
{
	Token tok= tz.gettoken ();
	switch (tok.type () )
	{
	case TypeAF:
		parsecomma (tz);
		tok= tz.gettoken ();
		if (tok.type () != TypeAFp)
			throw "Invalid operand for EX";
		gencode (0x08);
		cout << "08\tEX AF, AF'" << endl;
		break;
	case TypeDE:
		parsecomma (tz);
		tok= tz.gettoken ();
		if (tok.type () != TypeHL)
			throw "Invalid operand for EX";
		gencode (0xEB);
		cout << "EB\tEX DE, HL" << endl;
		break;
	case TypeOpen:
		tok= tz.gettoken ();
		if (tok.type () != TypeSP)
			throw "Invalid operand for EX";
		parseclose (tz);
		parsecomma (tz);
		tok= tz.gettoken ();
		switch (tok.type () )
		{
		case TypeHL:
			gencode (0xE3);
			cout << "E3\tEX (SP), HL" << endl;
			break;
		case TypeIX:
			gencode (prefixIX);
			gencode (0xE3);
			cout << "DD E3\tEX (SP), IX" << endl;
			break;
		case TypeIY:
			gencode (prefixIY);
			gencode (0xE3);
			cout << "FD E3\tEX (SP), IY" << endl;
			break;
		default:
			throw "Invalid operand for EX";
		}
		break;
	default:
		throw "Invalid operand for EX";
	}
	checkendline (tz);
}

void Asm::parseIN (Tokenizer & tz)
{
	Token tok= tz.gettoken ();
	byte code;
	switch (tok.type () )
	{
	case TypeB:
		code= 0x40; break;
	case TypeC:
		code= 0x48; break;
	case TypeD:
		code= 0x50; break;
	case TypeE:
		code= 0x58; break;
	case TypeH:
		code= 0x60; break;
	case TypeL:
		code= 0x68; break;
	case TypeA:
		parsecomma (tz);
		parseopen (tz);
		tok= tz.gettoken ();
		if (tok.type () == TypeC)
		{
			parseclose (tz);
			gencode (0xED);
			gencode (0x78);
			cout << "ED 78\tIN A, (C)" << endl;
		}
		else
		{
			//address addr= getvalue0 (tok);
			address addr= parseexpr (false, tok, tz);
			byte b= static_cast <byte> (addr);
			gencode (0xDB);
			gencode (b);
			cout << "DB " << hex2 << int (b) <<
				"\tIN A, (n)" << endl;
			parseclose (tz);
		}
		checkendline (tz);
		return;
	default:
		throw "Invalid operand";
	}
	parsecomma (tz);
	parseopen (tz);
	parseC (tz);
	parseclose (tz);
	checkendline (tz);
	gencode (0xED);
	gencode (code);
	cout << "ED " << hex2 << int (code) << endl;
}

void Asm::parseOUT (Tokenizer & tz)
{
	parseopen (tz);
	Token tok= tz.gettoken ();
	if (tok.type () != TypeC)
	{
		//address addr= getvalue0 (tok);
		address addr= parseexpr (false, tok, tz);
		byte b= static_cast <byte> (addr);
		parseclose (tz);
		parsecomma (tz);
		parseA (tz);
		checkendline (tz);
		gencode (0xD3);
		gencode (b);
		cout << "D3 " << hex2 << int (b) <<
			"\tOUT (n), A" << endl;
		return;
	}
	parseclose (tz);
	parsecomma (tz);
	tok= tz.gettoken ();
	byte code;
	switch (tok.type () )
	{
	case TypeA:
		code= 0x79; break;
	case TypeB:
		code= 0x41; break;
	case TypeC:
		code= 0x49; break;
	case TypeD:
		code= 0x51; break;
	case TypeE:
		code= 0x59; break;
	case TypeH:
		code= 0x61; break;
	case TypeL:
		code= 0x69; break;
	default:
		throw "Invalid operand";
	}
	checkendline (tz);
	gencode (0xED);
	gencode (code);
	cout << "ED " << hex2 << int (code) << endl;
}

void Asm::dobit (Tokenizer & tz, byte basecode)
{
	Token tok= tz.gettoken ();
	//address addr= getvalue0 (tok);
	address addr= parseexpr (false, tok, tz);
	if (addr > 7)
		throw "Bit position out of range";
	parsecomma (tz);
	dobyteparamCB (tz, basecode + (addr << 3) );
}

void Asm::parseBIT (Tokenizer & tz)
{
	dobit (tz, 0x40);
}

void Asm::parseRES (Tokenizer & tz)
{
	dobit (tz, 0x80);
}

void Asm::parseSET (Tokenizer & tz)
{
	dobit (tz, 0xC0);
}

void Asm::parseDEFB (Tokenizer & tz)
{
	for (;;)
	{
		Token tok= tz.gettoken ();
		switch (tok.type () )
		{
		case TypeNumber:
			gencode (tok.num () & 0xFF);
			break;
		//case TypeIdentifier:
		//	gencode (getvalue0 (tok.str () ) & 0xFF);
		//	break;
		case TypeLiteral:
			{
				const std::string & str= tok.str ();
				const std::string::size_type l= str.size ();
				if (l == 1)
				{
					// Admit expressions like 'E' + 80H
					gencode (parseexpr (false, tok, tz) );
					break;
				}
				for (std::string::size_type i= 0; i < l; ++i)
				{
					gencode (str [i] );
				}
			}
			break;
		default:
			gencode (parseexpr (false, tok, tz) );
		}
		tok= tz.gettoken ();
		if (tok.type () == TypeEndLine)
			break;
		if (tok.type () != TypeComma)
			throw "Error de sintaxis";
	}
}

void Asm::parseDEFW (Tokenizer & tz)
{
	for (;;)
	{
		Token tok= tz.gettoken ();
		//gencodeword (getvalue0 (tok) );
		gencodeword (parseexpr (false, tok, tz) );
		tok= tz.gettoken ();
		if (tok.type () == TypeEndLine)
			break;
		if (tok.type () != TypeComma)
			throw "Error de sintaxis";
	}
}

void Asm::parseDEFS (Tokenizer & tz)
{
	Token tok= tz.gettoken ();
	//address addr= getvalue (tok);
	address addr= parseexpr (true, tok, tz);
	byte value= 0;
	tok= tz.gettoken ();
	switch (tok.type () )
	{
	case TypeEndLine:
		// Nothing to do
		break;
	case TypeComma:
		tok= tz.gettoken ();
		{
			//address addr= getvalue0 (tok);
			address addr= parseexpr (false, tok, tz);
			checkendline (tz);
			value= static_cast <byte> (addr);
		}
		break;
	default:
		throw "Syntax error";
	}
	for (address i= 0; i < addr; ++i)
		gencode (value);
}

void Asm::emitobject (std::ostream & out)
{
	cout << "Emitiendo codigo desde " << hex4 << minused <<
		" hasta " << hex4 << maxused << endl;

	for (int i= minused; i <= maxused; ++i)
	{
		out.put (mem [i] );
	}
}

void Asm::emitplus3dos (std::ostream & out)
{
	cout << "Emitiendo PLUS3DOS desde " << hex4 << minused <<
		" hasta " << hex4 << maxused << endl;

	// PLUS3DOS haeader.

	byte plus3 [128]= {
		'P', 'L', 'U', 'S', '3', 'D', 'O', 'S', // Identifier.
		0x1A, // CP/M EOF.
		1,    // Issue number.
		0,    // Version number.
	};
	address codesize= maxused - minused + 1;
	size_t filesize= codesize + 128;
	// Size of file including code and header.
	plus3 [11]= filesize & 0xFF;
	plus3 [12]= (filesize >> 8) & 0xFF;
	plus3 [13]= (filesize >> 16) & 0xFF;
	plus3 [14]= (filesize >> 24) & 0xFF;
	plus3 [15]= 3; // Type: code.
	// Size of code.
	plus3 [16]= codesize & 0xFF;
	plus3 [17]= (codesize >> 8) & 0xFF;
	// Start address.
	plus3 [18]= minused & 0xFF;
	plus3 [19]= (minused >> 8) & 0xFF;
	// Don't know if used for something.
	plus3 [20]= 0x80;
	plus3 [21]= 0x80;
	// Checksum
	byte check= 0;
	for (int i= 0; i < 127; ++i)
		check+= plus3 [i];
	plus3 [127]= check;

	// Write the file.

	out.write (reinterpret_cast <char *> (plus3), sizeof (plus3) );
	out.write (reinterpret_cast <char *> (mem + minused), codesize);
	size_t round= 128 - (codesize % 128);
	if (round != 128)
	{
		char aux [128]= {};
		out.write (aux, round);
	}
	if (! out)
		throw "Error writing";
}

void Asm::emittap (std::ostream & out, const std::string & filename)
{
	cout << "Emitiendo TAP desde " << hex4 << minused <<
		" hasta " << hex4 << maxused << endl;

	address codesize= maxused - minused + 1;
	byte block1 [21]= {
		0x13, 0x00, // Length of block: 17 bytes + flag + checksum.
		0x00,       // Flag: 00 -> header
		0x03,       // Type: code block.
	};
	std::string::size_type l= filename.size ();
	if (l > 10)
		l= 10;
	for (std::string::size_type i= 0; i < 10; ++i)
		block1 [4 + i]= i < l ? filename [i] : ' ';
	// Length of the code block.
	block1 [14]= codesize & 0xFF;
	block1 [15]= codesize >> 8;
	// Start of the code block.
	block1 [16]= minused & 0xFF;
	block1 [17]= minused >> 8;
	// Parameter 2: 32768 in a code block.
	block1 [18]= 0x00;
	block1 [19]= 0x80;
	// Checksum
	byte check= block1 [2]; // Flag byte included.
	for (int i= 3; i < 20; ++i)
		check^= block1 [i];
	block1 [20]= check;
	byte headblock2 [3]= { };
	address blocksize= codesize + 2; // Code + flag + checksum.
	headblock2 [0]= blocksize & 0xFF;
	headblock2 [1]= blocksize >> 8;
	headblock2 [2]= 0xFF; // Flag: data block.
	// Checksum of block 2.
	check= 0xFF; // Flag byte included.
	for (int i= minused; i < static_cast <int> (maxused) + codesize; ++i)
		check^= mem [i];
	out.write (reinterpret_cast <char *> (block1), sizeof (block1) );
	out.write (reinterpret_cast <char *> (headblock2),
		sizeof (headblock2) );
	out.write (reinterpret_cast <char *> (mem + minused), codesize);
	out.write (reinterpret_cast <char *> (& check), 1);
}

void Asm::emithex (std::ostream & out)
{
	cout << "Emitiendo Intel HEX desde " << hex4 << minused <<
		" hasta " << hex4 << maxused << endl;

	address end= maxused + 1;
	for (address i= minused; i < end; i+= 16)
	{
		address len= end - i;
		if (len > 16)
			len= 16;
		out << ':' << hex2 << len << hex4 << i << "00";
		byte sum= len + ( (i >> 8) & 0xFF) + i & 0xFF;
		for (address j= 0; j < len; ++j)
		{
			byte b= mem [i + j];
			out << hex2 << int (b);
			sum+= b;
		}
		out << hex2 << 0x100 - sum;
		out << "\r\n";
	}
	out << ":00000001FF\r\n";
}

void Asm::emitamsdos (std::ostream & out, const std::string & filename)
{
	cout << "Emitiendo Amsdos desde " << hex4 << minused <<
		" hasta " << hex4 << maxused << endl;

	byte amsdos [128]= { 0 };

	address codesize= maxused - minused + 1;
	//address len= codesize + 128;
	address len= codesize;

	// 00: User number, use 0.
	// 01-0F: filename, padded with 0.
	std::string::size_type l= filename.size ();
	if (l > 15)
		l= 15;
	for (std::string::size_type i= 0; i < l; ++i)
		amsdos [i + 1]= filename [i];
	for (std::string::size_type i= l; i < 15; ++i)
		amsdos [i + 1]= '\0';
	// 10: Block number, 0 in disk files.
	// 11: last block flag, 0 in disk files.
	amsdos [0x12]= 2; // File type: binary.
	// 13-14 // Length of block, 0 in disk files
	// 15-16: Load address.
	amsdos [0x15]= minused & 0xFF;
	amsdos [0x16]= minused >> 8;
	// 17: first block flag, 0 in disk files.
	// 18-19: logical length.
	amsdos [0x18]= len & 0xFF;
	amsdos [0x19]= len >> 8;
	// 1A-1B: Entry address.
	address entry= 0;
	if (hasentrypoint)
		entry= entrypoint;
	amsdos [0x1A]= entry & 0xFF;
	amsdos [0x1B]= entry >> 8;
	// 1C-3F: Unused.
	// 40-42: real length of file.
	amsdos [0x40]= len & 0xFF;
	amsdos [0x41]= (len >> 8) & 0xFF;
	amsdos [0x42]= 0;
	// 43-44 checksum of bytes 00-42
	address check= 0;
	for (int i= 0; i < 0x43; ++i)
		check+= amsdos [i];
	amsdos [0x43]= check & 0xFF;
	amsdos [0x44]= check >> 8;
	// 45-7F: Unused.

	out.write (reinterpret_cast <char *> (amsdos), sizeof (amsdos) );
	out.write (reinterpret_cast <char *> (mem + minused), codesize);
}

void Asm::emitprl (std::ostream & out, Asm & asmoff)
{
	static const std::string
		nosync ("PRL genration failed: out of sync");

	if (minused - base != asmoff.minused - asmoff.base)
		throw nosync;
	if (maxused - base != asmoff.maxused - asmoff.base)
		throw nosync;
	ushort len= maxused - minused + 1;
	ushort off= asmoff.base - base;

	// PRL header.

	byte prlhead [256]= { 0 };
	prlhead [1]= len & 0xFF;
	prlhead [2]= len >> 8;
	out.write (reinterpret_cast <char *> (prlhead), sizeof (prlhead) );
	ushort reloclen= (len + 7) / 8;
	byte * reloc= new byte [reloclen];
	memset (reloc, 0, reloclen);

	// Build relocation bitmap.
	for (ushort i= minused; i <= maxused; ++i)
	{
		byte b= mem [i];
		byte b2= asmoff.mem [i + off];
		if (b != b2)
		{
			if (b2 - b != off / 256)
			{
				cerr << "off= " << hex4 << off <<
					", b= " << hex2 << int (b) <<
					", b2= " << hex2 << int (b2) <<
					dec << endl;
				throw nosync;
			}
			ushort pos= i - minused;
			static const byte mask [8]= {
				0x80, 0x40, 0x20, 0x10,
				0x08, 0x04, 0x02, 0x01
			};
			reloc [pos / 8]|= mask [pos % 8];
		}
	}

	// Write code in position 0x0100
	out.write (reinterpret_cast <char *> (asmoff.mem + minused + off),
		len);
	// Write relocation bitmap.
	out.write (reinterpret_cast <char *> (reloc), reloclen);
}

void Asm::dumppublic (std::ostream & out)
{
	for (setpublic_t::iterator pit= setpublic.begin ();
		pit != setpublic.end ();
		++pit)
	{
		mapvar_t::iterator it= mapvar.find (* pit);
		if (it != mapvar.end () )
		{
			out << it->first << "\tEQU 0" << hex4 <<
				it->second << 'H' << endl;
		}
	}
}

void Asm::dumpsymbol (std::ostream & out)
{
	for (mapvar_t::iterator it= mapvar.begin ();
		it != mapvar.end ();
		++it)
	{
		out << it->first << "\tEQU 0" << hex4 << it->second << 'H'
			<< endl;
	}
}

// End of asm.cpp
