// asm.h
// Revision 26-jan-2004

#include <ostream>
#include <string>
#include <sstream>
#include <vector>
#include <map>
#include <set>
#include <deque>

#include "token.h"

class Asm {
public:
	Asm ();
	~Asm ();
	void setbase (address addr);
	void setentrypoint (address addr);
	void checkendline (Tokenizer & tz);
	void gencode (byte code);
	void gencodeword (address value);
	void setvalue (const std::string & var, address value);
	address getvalue (const std::string & var);
	address getvalue (const Token & tok);
	address getvalue0 (const std::string & var);
	address getvalue0 (const Token & tok);
	void parsevalue (bool required, Tokenizer & tz, address & result);
	void parseopen (bool required, Tokenizer & tz, address & result);
	void parsemuldiv (bool required, Tokenizer & tz, address & result);
	void parseplusmin (bool required, Tokenizer & tz, address & result);
	void parsenot (bool required, Tokenizer & tz, address & result);
	void parseand (bool required, Tokenizer & tz, address & result);
	void parseorxor (bool required, Tokenizer & tz, address & result);
	address parseexpr (bool required, const Token & tok, Tokenizer & tz);
	void parsecomma (Tokenizer & tz);
	void parseopen (Tokenizer & tz);
	void parseclose (Tokenizer & tz);
	void parseA (Tokenizer & tz);
	void parseC (Tokenizer & tz);
	void loadfile (const std::string & filename);
	void getvalidline ();
	void parseif (Tokenizer & tz);
	void parseelse ();
	void parseline (Tokenizer & tz);
	void dopass ();
	void processfile (const std::string & filename);
	void parsegeneric (Tokenizer & tz, Token tok);
	void parseorg (Tokenizer & tz);
	void parselabel (Tokenizer & tz, const std::string & name);
	void parsedesp (Tokenizer & tz, byte & desp);
	bool parsebyteparam (Tokenizer & tz, TypeToken tt,
		unsigned short & regcode,
		byte & prefix, bool & hasdesp, byte & desp);
	void dobyteparam (Tokenizer & tz, byte codereg, byte codein);
	void dobyteparamCB (Tokenizer & tz, byte codereg);
	void parseIM (Tokenizer & tz);
	void parseRST (Tokenizer & tz);
	void parseLDA (Tokenizer & tz);
	void parseLDsimple (Tokenizer & tz, unsigned short regcode);
	void parseLDdouble (Tokenizer & tz,
		unsigned short regcode, byte prefix);
	void parseLDSP (Tokenizer & tz);
	void parseLD (Tokenizer & tz);
	void parseCP (Tokenizer & tz);
	void parseAND (Tokenizer & tz);
	void parseOR (Tokenizer & tz);
	void parseXOR (Tokenizer & tz);
	void parseRL (Tokenizer & tz);
	void parseRLC (Tokenizer & tz);
	void parseRR (Tokenizer & tz);
	void parseRRC (Tokenizer & tz);
	void parseSLA (Tokenizer & tz);
	void parseSRA (Tokenizer & tz);
	void parseSRL (Tokenizer & tz);
	void parseSUB (Tokenizer & tz);
	void parseADDADCSBCHL (Tokenizer & tz, byte prefix, byte basecode);
	void parseADD (Tokenizer & tz);
	void parseADC (Tokenizer & tz);
	void parseSBC (Tokenizer & tz);
	void parsePUSHPOP (Tokenizer & tz, bool isPUSH);
	void parseCALL (Tokenizer & tz);
	void parseRET (Tokenizer & tz);
	void parseJP (Tokenizer & tz);
	void parseJR (Tokenizer & tz);
	void parseDJNZ (Tokenizer & tz);
	void parseINCDEC (Tokenizer & tz, bool isINC);
	void parseEX (Tokenizer & tz);
	void parseIN (Tokenizer & tz);
	void parseOUT (Tokenizer & tz);
	void dobit (Tokenizer & tz, byte basecode);
	void parseBIT (Tokenizer & tz);
	void parseRES (Tokenizer & tz);
	void parseSET (Tokenizer & tz);
	void parseDEFB (Tokenizer & tz);
	void parseDEFW (Tokenizer & tz);
	void parseDEFS (Tokenizer & tz);
	void parsePUBLIC (Tokenizer & tz);
	void parseEND (Tokenizer & tz);
	void emitobject (std::ostream & out);
	void emitplus3dos (std::ostream & out);
	void emittap (std::ostream & out, const std::string & filename);
	void emithex (std::ostream & out);
	void emitamsdos (std::ostream & out, const std::string & filename);
	void emitprl (std::ostream & out, Asm & asmoff);
	void dumppublic (std::ostream & out);
	void dumpsymbol (std::ostream & out);
private:
	byte mem [65536];
	address base;
	address current;
	address minused;
	address maxused;
	address entrypoint;
	bool hasentrypoint;
	size_t currentline;
	typedef std::map <std::string, address> mapvar_t;
	mapvar_t mapvar;
	typedef std::set <std::string> setpublic_t;
	setpublic_t setpublic;

	int pass;

	typedef std::vector <std::string> vline_t;
	vline_t vline;
	struct FileRef {
		std::string filename;
		FileRef (const std::string & name) :
			filename (name)
		{ }
	};
	std::vector <FileRef> vfileref;
	struct LineInfo {
		size_t linenum;
		size_t filenum;
		LineInfo (size_t line, size_t file) :
			linenum (line), filenum (file)
		{ }
	};
	std::vector <LineInfo> vlineinfo;
	size_t iflevel;
};

// End of asm.h
