// miasma.cpp
// Revision 26-jan-2004

#include "asm.h"

#include <fstream>
#include <iostream>
using std::cerr;
using std::endl;

enum ObjectFormat {
	ObjectRaw, ObjectHex, ObjectPrl,
	ObjectPlus3Dos, ObjectTap, ObjectAmsdos,
};

int miasma (int argc, char * * argv)
{
	ObjectFormat format= ObjectRaw;

	bool emitpublic= false;
	int argpos= 1;
	std::string headername;
	for ( ; argpos < argc; ++argpos)
	{
		const std::string arg (argv [argpos] );
		if (arg == "--hex")
			format= ObjectHex;
		else if (arg == "--prl")
			format= ObjectPrl;
		else if (arg == "--plus3dos")
			format= ObjectPlus3Dos;
		else if (arg == "--tap")
			format= ObjectTap;
		else if (arg == "--amsdos")
			format= ObjectAmsdos;
		else if (arg == "--public")
			emitpublic= true;
		else if (arg == "--name")
		{
			++argpos;
			if (argpos >= argc)
				throw "Option --name needs an argument";
			headername= argv [argpos];
		}
		else if (arg == "--")
		{
			++argpos;
			break;
		}
		else
			break;
	}

	if (argpos >= argc)
		throw "Falta fuente";
	char * filein= argv [argpos];
	++argpos;
	if (argpos >= argc)
		throw "Falta objeto";
	char * fileout= argv [argpos];
	++argpos;
	char * filesymbol= NULL;
	if (argpos < argc)
	{
		filesymbol= argv [argpos];
		++argpos;
		if (argpos < argc)
			cerr << "Argumentos extra ignorados" << endl;
	}

	if (headername.empty () )
		headername= fileout;

	Asm a;
	a.processfile (filein);

	std::ofstream out (fileout,
		std::ios::out | std::ios::binary);
	if (! out.is_open () )
		throw "Error al crear fichero objeto";

	switch (format)
	{
	case ObjectRaw:
		a.emitobject (out);
		break;
	case ObjectHex:
		a.emithex (out);
		break;
	case ObjectPrl:
		{
			Asm aoff;
			aoff.setbase (0x100);
			aoff.processfile (filein);
			a.emitprl (out, aoff);
		}
		break;
	case ObjectPlus3Dos:
		a.emitplus3dos (out);
		break;
	case ObjectTap:
		a.emittap (out, headername);
		break;
	case ObjectAmsdos:
		a.emitamsdos (out, headername);
	}
	out.close ();

	if (filesymbol)
	{
		std::ofstream sout (filesymbol);
		if (! sout.is_open () )
			throw "Error al crear fichero de smbolos";
		if (emitpublic)
			a.dumppublic (sout);
		else
			a.dumpsymbol (sout);
	}

	return 0;
}

int main (int argc, char * * argv)
{
	try
	{
		return miasma (argc, argv);
	}
	catch (const char * str)
	{
		cerr << "ERROR: " << str << endl;
		return 1;
	}
	catch (const std::string & str)
	{
		cerr << "ERROR: " << str << endl;
		return 1;
	}
}

// End of miasma.cpp
