// token.h
// Revision 13-jan-2004

#include <string>
#include <sstream>
#include <deque>

typedef unsigned short address;
typedef unsigned char byte;

enum TypeToken {
	TypeUndef= 0,

	TypeEndLine= 1,
	TypeComma= ',',
	TypeColon= ':',
	TypePlus= '+',
	TypeMinus= '-',
	TypeMult= '*',
	TypeDiv= '/',
	TypeEqual= '=',
	TypeOpen= '(',
	TypeClose= ')',
	TypeDollar= '$',

	// Valores
	TypeIdentifier= 0x100,
	TypeLiteral,
	TypeNumber,

	// Operadores.
	TypeMOD,
	TypeSHL,
	TypeSHR,
	TypeNOT,
	// AND, OR y XOR son nemnicos.

	// Directivas
	TypeDEFB,
	TypeDB,
	TypeDEFM,
	TypeDEFW,
	TypeDW,
	TypeDEFS,
	TypeDS,
	TypeEQU,
	TypeORG,
	TypeINCLUDE,
	TypeIF,
	TypeELSE,
	TypeENDIF,
	TypePUBLIC,
	TypeEND,

	// Nemonicos
	TypeADC,
	TypeADD,
	TypeAND,
	TypeBIT,
	TypeCALL,
	TypeCCF,
	TypeCP,
	TypeCPD,
	TypeCPDR,
	TypeCPI,
	TypeCPIR,
	TypeCPL,
	TypeDAA,
	TypeDEC,
	TypeDI,
	TypeDJNZ,
	TypeEI,
	TypeEX,
	TypeEXX,
	TypeHALT,
	TypeIM,
	TypeIN,
	TypeINC,
	TypeIND,
	TypeINDR,
	TypeINI,
	TypeINIR,
	TypeJP,
	TypeJR,
	TypeLD,
	TypeLDD,
	TypeLDDR,
	TypeLDI,
	TypeLDIR,
	TypeNEG,
	TypeNOP,
	TypeOR,
	TypeOTDR,
	TypeOTIR,
	TypeOUT,
	TypeOUTD,
	TypeOUTI,
	TypePOP,
	TypePUSH,
	TypeRES,
	TypeRET,
	TypeRETI,
	TypeRETN,
	TypeRL,
	TypeRLA,
	TypeRLC,
	TypeRLCA,
	TypeRLD,
	TypeRR,
	TypeRRA,
	TypeRRC,
	TypeRRCA,
	TypeRRD,
	TypeRST,
	TypeSBC,
	TypeSCF,
	TypeSET,
	TypeSLA,
	TypeSRA,
	TypeSRL,
	TypeSUB,
	TypeXOR,
	
	// Registers
	// C is listed as flag.
	TypeA,
	TypeAF,
	TypeAFp, // AF'
	TypeB,
	TypeBC,
	TypeD,
	TypeE,
	TypeDE,
	TypeH,
	TypeL,
	TypeHL,
	TypeSP,
	TypeIX,
	TypeIY,
	TypeI,
	TypeR,

	// Flags
	TypeNZ,
	TypeZ,
	TypeNC,
	TypeC,
	TypePO,
	TypePE,
	TypeP,
	TypeM,
};

class Token {
public:
	Token ();
	Token (TypeToken t);
	Token (address n);
	Token (TypeToken t, const std::string & s);
	TypeToken type () const { return t; }
	std::string str () const;
	address num () const { return number; }
private:
	TypeToken t;
	std::string s;
	address number;
};

class Tokenizer {
public:
	Tokenizer (const std::string & line);
	~Tokenizer ();
	void putback (const Token & tok);
	Token gettoken ();
	std::string getincludefile ();
private:
	//std::istringstream iss;
	Tokenizer (const Tokenizer &); // Forbidden.
	void operator = (const Tokenizer &); // Forbidden.

	typedef std::deque <Token> tokenlist_t;
	tokenlist_t tokenlist;
	tokenlist_t::iterator current;

	bool hassaved;
	Token saved;

	Token parsetoken (std::istringstream & iss);
	std::string parseincludefile (std::istringstream & iss);
};

// End of token.h
